use std::time::SystemTime;

use crate::util::strip_path_prefix;



/// The events we are worried about.
///
/// These will be sent as messages in a channel.
pub enum Event
{
   /// A file has been created.
   New
   {
      /// The path to the file.
      path: std::path::PathBuf,

      /// The modification time of the file.
      mod_time: SystemTime
   },

   /// A file has been modified.
   Modify
   {
      /// The path to the file.
      path: std::path::PathBuf,

      /// The modification time of the file.
      mod_time: SystemTime
   },

   /// A file has been renamed, and it stayed in a monitored directory.
   Rename
   {
      /// The path of the file that was renamed.
      from: std::path::PathBuf,

      /// The path that the file was renamed to.
      to: std::path::PathBuf,

      /// The modification time of the file.
      mod_time: SystemTime
   },

   /// A file has been removed.
   Delete
   {
      /// The path of the removed file.
      path: std::path::PathBuf
   }
}



impl Event
{
   /// Take all the paths in the Event and makes them relative
   /// to the base_dir directory.
   pub fn make_paths_relative(&mut self, base_dir: &std::path::Path)
   {
      match self
      {
         Event::New { path, mod_time: _ } =>
         {
            *path = strip_path_prefix(&path, base_dir);
         }
         Event::Modify { path, mod_time: _ } =>
         {
            *path = strip_path_prefix(&path, base_dir);
         }
         Event::Rename { from,
                         to,
                         mod_time: _ } =>
         {
            *from = strip_path_prefix(&from, base_dir);
            *to = strip_path_prefix(&to, base_dir);
         }
         Event::Delete { path } =>
         {
            *path = strip_path_prefix(&path, base_dir);
         }
      }
   }
}


impl std::fmt::Display for Event
{
   fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result
   {
      match self
      {
         Event::New { path, mod_time: _ } =>
         {
            write!(f, "[NEW] {}", path.display())
         }
         Event::Modify { path, mod_time: _ } =>
         {
            write!(f, "[MOD] {}", path.display())
         }
         Event::Rename { from,
                         to,
                         mod_time: _ } =>
         {
            writeln!(f, "[DEL] {}", from.display())?;
            write!(f, "[NEW] {}", to.display())
         }
         Event::Delete { path } =>
         {
            write!(f, "[DEL] {}", path.display())
         }
      }
   }
}
