//! Rust challenge: FileMonitor



mod dir_monitor;
mod event;
mod monitored_files;
mod notify_sender;
mod util;



use clap::Parser;
use tokio::sync::watch;

use crate::dir_monitor::DirMonitor;



#[derive(Parser)]
#[command(version)]
/// A file monitor progam as part of a rust challenge.
///
/// This is a rust challenge to create a file monitoring program that will
/// watch a single directory and its sub directories for changes to the
/// files.
struct Options
{
   #[arg(short = 'd', default_value = "~/inbox")]
   /// The path for the inbox to monitor.
   inbox_dir: String
}



/// Creates the given directory, and its parents, if it does not exist.
fn create_directory(dir: &str) -> Option<String>
{
   let mut inbox_dir = String::from(dir);
   if inbox_dir.starts_with("~")
   {
      if let Some(path) = home::home_dir()
      {
         inbox_dir = inbox_dir.replace("~", &path.display().to_string());
      }
   }

   match std::fs::create_dir_all(inbox_dir.clone())
   {
      Ok(_) => Some(inbox_dir),

      Err(e) =>
      {
         eprintln!("Creating directory error:  {}", e);
         None
      }
   }
}

/// Asynchronously listen for CTRL-C to be pressed.
///
/// This will send a false message across the channel
/// when it detects the termination signal.
async fn listen_for_termination(sender: watch::Sender<bool>)
{
   tokio::signal::ctrl_c().await.unwrap();

   // CTRL-C was pressed so send a message to everyone that is
   // watching that the program is to be terminated. Then close the
   // sender.
   sender.send_replace(false);
   sender.closed().await;
}


/// The usual starting point of your project.
#[tokio::main]
async fn main()
{
   let options: Options = Options::parse();

   if let Some(inbox_dir) = create_directory(&options.inbox_dir)
   {
      // This is our keep running channel. False will be sent
      // when the application is meant to be closed.
      let (sender, receiver) = watch::channel(true);

      let mut directory_monitor: DirMonitor = DirMonitor::new(&inbox_dir);

      // Program workflow step 1.
      // Async is not needed here. This can be done syncronously because
      // file systems are not async and the spawn_blocking blocking overhead
      // isn't needed.
      directory_monitor.scan();
      directory_monitor.print_monitored_files();

      // Run our directory monitor.
      let mon_task = tokio::spawn(async move {
         // Program workflow step 2 (terminated by step 3).
         // This can be done async because we are waiting on
         // events to be loaded from the notify library and
         // we want this to run continously and not block the
         // program. This could also be done on a seperate thread,
         // preferably a green thread, but a fiber is sufficient
         // and less resource intensive.
         directory_monitor.monitor(receiver).await;

         // Program workflow step 4.
         // Async is not needed here. This can be done
         // syncronously because it is the final/clean up
         // step of the program.
         //
         // This is done here because we have moved the directory
         // monitor inside this task.
         directory_monitor.print_monitored_files();
      });

      // Run until Ctrl-C is pressed.
      // Once it is pressed it will send a message to stop.
      let _term_task = tokio::spawn(async move {
         // Program workflow step 3.
         // Async with message passing so it can run and listen
         // for the termination signal. It uses a watch channel
         // because any task that will need to know when the
         // program is to end can then catch the change.
         listen_for_termination(sender).await;
      });

      // Only need to wait for the monitor task to finish because it will only
      // finish once the termination task has finished.
      match mon_task.await
      {
         Ok(_) =>
         {}
         Err(e) =>
         {
            eprintln!("Error during monitoring task:  {}", e);
         }
      }
   }
}
