use std::collections::HashMap;
use std::path::{Path, PathBuf};
use std::time::SystemTime;

use chrono::{DateTime, Local};

use crate::event::Event;



/// The files currently being monitored and their last modification time.
pub struct MonitoredFiles
{
   /// A mapping of a Path to the last known modification time.
   files: HashMap<PathBuf, SystemTime>
}



impl MonitoredFiles
{
   /// Create a new empty set of monitored files.
   pub fn new() -> Self
   {
      MonitoredFiles { files: HashMap::new() }
   }

   /// Process an event and change the monitored files.
   pub fn process_event(&mut self, event: Event)
   {
      match event
      {
         Event::New { path, mod_time } =>
         {
            self.add(&path, mod_time);
         }

         Event::Modify { path, mod_time } =>
         {
            self.modify(&path, mod_time);
         }

         Event::Rename { from, to, mod_time } =>
         {
            self.remove(&from);
            self.add(&to, mod_time);
         }

         Event::Delete { path } =>
         {
            self.remove(&path);
         }
      }
   }

   /// Add a file that is to be monitored..
   pub fn add(&mut self, path: &Path, mod_time: SystemTime)
   {
      self.files.insert(PathBuf::from(path), mod_time);
   }

   /// Modify the stored modifcation time of a monitored file.
   pub fn modify(&mut self, path: &Path, time: SystemTime)
   {
      if let Some(entry) = self.files.get_mut(path)
      {
         *entry = time;
      }
   }

   /// Remove a file from being monitored.
   pub fn remove(&mut self, path: &Path)
   {
      self.files.remove(path);
   }
}


impl std::fmt::Display for MonitoredFiles
{
   fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result
   {
      let mut it = self.files.iter().peekable();
      while let Some((path, mod_time)) = it.next()
      {
         let date_time: DateTime<Local> = DateTime::from(*mod_time);

         if !it.peek().is_none()
         {
            writeln!(f,
                     "[{}] {}",
                     date_time.format("%m/%d/%Y %H:%M"),
                     path.display())?;
         }
         else
         {
            write!(f,
                   "[{}] {}",
                   date_time.format("%m/%d/%Y %H:%M"),
                   path.display())?;
         }
      }

      Ok(())
   }
}
