use std::path::PathBuf;

use crate::util::get_file_mod_time;



/// NotifySender is a newtype idiom around the tokio mpsc Sender.
///
/// This was done to allow the tokio mpsc Sender to be a Notify crate
/// EventHandler.
pub struct NotifySender
{
   /// The tokio runtime to spawn event handling tasks on.
   tokio_runtime: tokio::runtime::Handle,

   /// The channel to send event messages on.
   sender: tokio::sync::mpsc::Sender<crate::event::Event>
}



impl NotifySender
{
   /// Create a new type that wraps the tokio mpsc Sender.
   pub fn new(tokio_runtime: tokio::runtime::Handle,
              sender: tokio::sync::mpsc::Sender<crate::event::Event>)
              -> Self
   {
      NotifySender { tokio_runtime,
                     sender }
   }
}

impl notify::EventHandler for NotifySender
{
   fn handle_event(&mut self, event: notify::Result<notify::Event>)
   {
      match event
      {
         Ok(notify_event) =>
         {
            // Process the notify event and turn it into an event
            // our crate is interested in, one of our Events. If
            // we get a valid one then create a tokio task that will
            // send the event out as a message.
            if let Some(monitor_event) = process_notify_event(notify_event)
            {
               let notify_sender = self.sender.clone();

               self.tokio_runtime.spawn(async move {
                                    if let Err(e) =
                                       notify_sender.send(monitor_event).await
                                    {
                                       eprintln!("Notify EventHandler:  {}", e);
                                    }
                                 });
            }
         }

         Err(e) =>
         {
            eprintln!("Notify crate error:  {}", e);
         }
      }
   }
}

/// Processes the notify crate events into events that our program cares about.
fn process_notify_event(event: notify::Event) -> Option<crate::event::Event>
{
   match event.kind
   {
      notify::EventKind::Create(kind) =>
      {
         process_create_event(kind, event.paths)
      }

      notify::EventKind::Modify(kind) =>
      {
         process_modify_event(kind, event.paths)
      }

      notify::EventKind::Remove(kind) =>
      {
         process_remove_event(kind, event.paths)
      }

      _ => None
   }
}

/// Process the Create event into a New event.
fn process_create_event(event: notify::event::CreateKind, paths: Vec<PathBuf>)
                        -> Option<crate::event::Event>
{
   if event == notify::event::CreateKind::File
   {
      let mod_time: std::time::SystemTime = match get_file_mod_time(&paths[0])
      {
         Ok(time) => time,

         Err(e) =>
         {
            eprintln!("Unable to open file to retrieve modification time:  {}",
                      e);
            std::time::SystemTime::now()
         }
      };

      return Some(crate::event::Event::New { path: paths[0].clone(),
                                             mod_time: mod_time });
   }

   None
}

/// Process the modify event into either a Modify event or a Rename event.
fn process_modify_event(event: notify::event::ModifyKind, paths: Vec<PathBuf>)
                        -> Option<crate::event::Event>
{
   match event
   {
      // This is just handling the events. On my system they came across as Any
      // for the MetadataKind, but the documentation seemed to hint that
      // modification times should be on write events. So for the sake of
      // being done without testing this on different platforms the Any
      // modification event will trigger a time lookup.
      notify::event::ModifyKind::Any |
      notify::event::ModifyKind::Metadata(_) =>
      {
         let mod_time: std::time::SystemTime = match get_file_mod_time(&paths
                                                                          [0])
         {
            Ok(time) => time,

            Err(e) =>
            {
               eprintln!("Unable to open file to retrieve modification time:  {}",
                                    e);
               std::time::SystemTime::now()
            }
         };

         Some(crate::event::Event::Modify { path: paths[0].clone(),
                                            mod_time: mod_time })
      }

      // Handling a rename event so that during testing when a file is
      // moved it is still tracked.
      notify::event::ModifyKind::Name(mode) =>
      {
         match mode
         {
            // This is for when a file is renamed and both the to and from
            // are in the monitored directories.
            notify::event::RenameMode::Both =>
            {
               let mod_time: std::time::SystemTime =
                  match get_file_mod_time(&paths[1])
                  {
                     Ok(time) => time,

                     Err(e) =>
                     {
                        eprintln!("Unable to open file to retrieve \
                                   modification time:  {}",
                                  e);
                        std::time::SystemTime::now()
                     }
                  };

               Some(crate::event::Event::Rename { from: paths[0].clone(),
                                                  to: paths[1].clone(),
                                                  mod_time: mod_time })
            }

            // This is for when a file is renamed and only the from
            // file is in the monitored directories.
            notify::event::RenameMode::From =>
            {
               Some(crate::event::Event::Delete { path: paths[0].clone() })
            }

            // This is for when a file is renamed and both the to file
            // is in the monitored directories.
            notify::event::RenameMode::To =>
            {
               let mod_time: std::time::SystemTime =
                  match get_file_mod_time(&paths[0])
                  {
                     Ok(time) => time,

                     Err(e) =>
                     {
                        eprintln!("Unable to open file to retrieve \
                                   modification time:  {}",
                                  e);
                        std::time::SystemTime::now()
                     }
                  };

               Some(crate::event::Event::New { path: paths[0].clone(),
                                               mod_time: mod_time })
            }

            _ => None
         }
      }

      _ => None
   }
}

/// Process the remove event into a Delete event.
fn process_remove_event(event: notify::event::RemoveKind, paths: Vec<PathBuf>)
                        -> Option<crate::event::Event>
{
   if event == notify::event::RemoveKind::File
   {
      return Some(crate::event::Event::Delete { path: paths[0].clone() });
   }

   None
}
